import os.path, sys

if __name__ == "__main__":

    if __package__ == "":
        # if the package is run directly with the folder, add
        # the parent folder as part of PYTHONPATH for the
        # current execution environment so the uninstalled
        # "scraparser" package is discoverable.
        sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))

    from .cli import entry_point
    #  pylint: disable=no-value-for-parameter,unexpected-keyword-arg
    entry_point(obj={})



import click

import logging
from logging.handlers import RotatingFileHandler

@click.group()
@click.option('--debug/--no-debug', default=False)
@click.pass_context
def entry_point(ctx, debug):
    loglevel = logging.WARNING
    if debug:
        loglevel = logging.DEBUG

    # Declare logger with file logging
    logger = logging.getLogger('main')
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    fh = RotatingFileHandler('scraparser.log',
            maxBytes=4000000,
            backupCount=5)
    fh.setLevel(loglevel)
    fh.setFormatter(formatter)
    logger.addHandler(fh)



import os
import re
import sys

import camelot
import click
import pandas as pd

from . import click_helpers


def __pdf_tables_to_dataframe(filepath, headers=False, pages='all', is_valid_table=False, sanitize=False, offset=0, limit=-1):
    """Read all tables in a given PDF then concat all into one dataframe"""

    # Read all tables in a PDF to frames
    tables = camelot.read_pdf(filepath=filepath, pages=pages)
    frames = []

    # define a header sanitier
    sanitize_header = lambda a: a
    if sanitize != False:
        sanitize_header = lambda d: dict(map(lambda kv: (kv[0], sanitize(kv[1])), d.items()))

    firstHeaders = sanitize_header(tables[offset].df.iloc[0].to_dict())

    if is_valid_table == False:
        # If no validator function, set it to check if the first headers are identical
        is_valid_table = lambda a: sanitize_header(a.iloc[0].to_dict()) == firstHeaders

    # determine the end of loop
    n = offset + limit
    if n < offset or n > tables.n:
        n = tables.n

    # loop through all tables, determine if they are valid for concat,
    # then put it into the "frames" slice for concating
    for i in range(offset, n):
        if is_valid_table(tables[i].df):
            frames.append(tables[i].df[1:])

    # assign the first table header to be the overall header
    # if none is provided
    if headers == False:
        headers = firstHeaders

    # concat all data frames into a single one
    return pd.concat(frames, ignore_index=True, sort=False).rename(columns=headers)


def __sanitize_items_in(df, sanitize):
    """Sanitize the dataframe values if it has EOL or space in it"""
    # Sanitize the dataframe values
    for i, row in df.iterrows():
        for key, value in row.iteritems():
            if type(value) != str:
                continue
            if "\n" in value or " " in value:
                df.at[i, key] = sanitize(value)


@click.command()
@click.option('--leave-space', default=False, is_flag=True, help="Leave the space in scraped table cells in sanitation process.")
@click.option('--headers', default=False, help="Specify the header names, separated by commas, for replacing the overall table. Use delimited \"\\,\" if you need to use actual comma.")
@click.option('--pages', default='all', type=str, help="Specifies the page range to scan. Can be comma separated page number, range like \"2-4\" or \"10-end\". Read \"3.5.2  Specify page numbers\" in camelot documentation.")
@click.option('--offset', default=0, type=int, help="The number of table(s) to skip before the start of scanning.")
@click.option('--limit', default=-1, type=int, help="The total number of table(s) to scan.")
@click_helpers.do_each(outer_keyword='files', inner_keyword='filename')
@click_helpers.print_output('--print-filename', 'print_filename', default=True, help="Print the result CSV filename to STDOUT.")
@click_helpers.dataframe_to_csv(keep_input_keyword=True, force_in_place=True)
@click.pass_context
def parse_pdf_to_csv(
    ctx,
    leave_space,
    headers,
    pages,
    offset,
    limit,
    filename,
):
    """Parse the PDF file and join all the tables"""

    # Define string sanitize
    sanitize = lambda a: a.replace("\r", "").replace("\n", "").replace(" ", "")
    if leave_space:
        sanitize = lambda a: a.replace("\r", "").replace("\n", "")

    if headers != False:
        # Treat the headers input as comma separated values
        headers_list = list(map(lambda a: a.replace("\\,", ","), re.split(r'(?<!\\),', headers)))
        headers_dict = dict(enumerate(headers_list))
    else:
        headers_dict = False

    # concat all tables, which passes the "is_valid_table" test
    # into a single dataframe.
    df = __pdf_tables_to_dataframe(
        filename,
        headers=headers_dict,
        pages=pages,
        #is_valid_table=lambda df: df[0][0] in ["個案\n編號", "Case \nno."],
        sanitize=sanitize,
        limit=limit,
        offset=offset,
    )

    # Sanitize the dataframe values
    __sanitize_items_in(df, sanitize)

    # Write to CSV file
    return df



from .entry_point import entry_point
from .fix import fix
from .googlesheet import googlesheet
from .parse_pdf import parse_pdf_to_csv
from .scrap import scrap
from .sort import sort

commands = [
    fix,
    googlesheet,
    parse_pdf_to_csv,
    scrap,
    sort,
]

for command in commands:
    entry_point.add_command(command)



import sys

import click
import pandas as pd

from . import click_helpers


@click.command()
@click.option('--sort-as-number', default=True, is_flag=True, help="To treat the column like number for sorting. Default True")
@click.option('--column', default=0, type=int, help="The column number to sort with. Count starting with 0. Default 0")
@click_helpers.do_each(outer_keyword='files', inner_keyword='filename')
@click_helpers.print_output('--print_filename', default=True, help="Print the result CSV filename to STDOUT.")
@click_helpers.csv_to_dataframe(outer_keyword="filename", inner_keyword="df", keep_outer_keyword=True)
@click_helpers.dataframe_to_csv(input_keyword="filename")
@click.pass_context
def sort(ctx, column, sort_as_number, df):
    """Sort the data by the specified column. By default sort by the first column."""
    column_index = df.keys()[0]
    if sort_as_number:
        df[column_index] = pd.to_numeric(df[column_index])
    df.sort_values(by=column_index, inplace=True)
    return df



import inspect
import sys
import os
from functools import wraps

import click
import pandas as pd


def do_each(outer_keyword="inputs", inner_keyword="input", keep_outer_keyword=False):
    """
    Consumes the keyword argument named as the `outer_keyword`,
    loop through them to invoke the inner function with the string
    remapped to the `inner_keyword` name.
    """
    def decorator(fn):

        @click.argument(outer_keyword, nargs=-1)
        @click.pass_context
        @wraps(fn)
        def wrapper(ctx, *args, **kwargs):
            if outer_keyword not in kwargs:
                # TODO: throw some exception
                sys.exit(1)
            inputs = kwargs[outer_keyword]

            if len(inputs) == 0:
                inputs = sys.stdin

            results = [] # TODO: Rewrite as iterator
            for input in inputs:
                input = input.rstrip("\r\n")
                innerKwargs = kwargs.copy()
                innerKwargs[inner_keyword] = input
                if not keep_outer_keyword and inner_keyword != outer_keyword:
                    del innerKwargs[outer_keyword]
                results.append(ctx.invoke(fn, *args, **innerKwargs))
            return results

        return wrapper
    return decorator


def csv_to_dataframe(outer_keyword="filename", inner_keyword="df", keep_outer_keyword=False):
    """
    Consumes the keyword argument named as the `outer_keyword` as if
    it is filename, open the file, convert to panda.Dataframe, then
    pass the dataframe to the inner function as the keyword argument
    named by `inner_keyword`.
    """
    def decorator(fn):

        @click.pass_context
        @wraps(fn)
        def wrapper(ctx, *args, **kwargs):

            if outer_keyword not in kwargs:
                # TODO: throw some exception
                sys.exit(1)
            filename = kwargs[outer_keyword]

            innerKwargs = kwargs.copy()
            innerKwargs[inner_keyword] = pd.read_csv(filename)

            if not keep_outer_keyword and inner_keyword != outer_keyword:
                del innerKwargs[outer_keyword]

            return ctx.invoke(fn, *args, **innerKwargs)
        return wrapper
    return decorator


def dataframe_to_csv(input_keyword="filename", default_filename="output.csv", force_in_place=False, keep_input_keyword=False):
    """
    Consumes the keyword argument named, as specified by the `input_keyword`
    then, with reference to `in_place` and `default_filename`, output the dataframe
    to the default filename or the input filename.

    Also check the `print_filename` flag to see if it should print the filename to STDOUT.
    """
    def decorator(fn):

        @click.option('--in-place', default=False, is_flag=True, help="Modify the file in-place.")
        @click.pass_context
        @wraps(fn)
        def wrapper(ctx, in_place=False, *args, **kwargs):

            if input_keyword not in kwargs:
                # TODO: throw some exception
                sys.exit(1)
            filename = kwargs[input_keyword]

            # Pass everything to the inner function
            innerKwargs = kwargs.copy()
            if not keep_input_keyword:
                del innerKwargs[input_keyword]
            df = ctx.invoke(fn, *args, **innerKwargs)

            # Save output of inner function (supposed DataFrame)
            # to the output csv file.
            if in_place or force_in_place:
                filename_basename = os.path.basename(filename)
                filename_basic, filename_ext = os.path.splitext(filename_basename)
                if filename_ext.lower() != ".csv":
                    # rename the file to csv, if needed.
                    filename = os.path.join(
                        os.path.dirname(os.path.abspath(filename)),
                        filename_basic + ".csv",
                    )
            else:
                filename = default_filename

            df.to_csv(filename, index=False)
            return filename

        return wrapper
    return decorator


def post_hook(
    callback=None,
    flag_name="--run-post-hook",
    flag_keyword="run_post_hook",
    keep_flag_keyword=False,
    **decorator_kwargs,
):
    """
    Run the wrapped function anyway. If the specified `flag_name` specified is True,
    then run `callback(ctx, *args, **kwargs)` and discard its output. Return the
    original wrapped function output.

    The `flag_keyword` is for mapping the click option to keyword variables.
    Default mapped variable name is "run_post_hok". You may override it when needed.
    """
    def decorator(fn):

        @click.option(flag_name, flag_keyword, is_flag=True, **decorator_kwargs)
        @click.pass_context
        @wraps(fn)
        def wrapper(ctx, *args, **kwargs):
            if callback is None:
                # TODO: throw some exception
                sys.exit(1)
            if flag_keyword not in kwargs:
                # TODO: throw some exception
                sys.exit(1)

            should_print = kwargs[flag_keyword]
            if not keep_flag_keyword:
                del kwargs[flag_keyword]

            # Pass everything to the inner function
            output = ctx.invoke(fn, *args, **kwargs)

            # If flagged to print output, print it.
            if should_print:
                callback(ctx, *args, **kwargs)
            return output

        return wrapper
    return decorator


def print_output(flag_name="--print", flag_keyword="print", keep_flag_keyword=False, **decorator_kwargs):
    """
    Run the wrapped function and print any output if the specified `flag_name`
    is specified in the CLI as True (or not set and default True).

    The `flag_keyword` is for mapping the click option to keyword variables.
    Default mapped variable name is "print". You may override it when needed.
    """
    def decorator(fn):

        @click.option(flag_name, flag_keyword, is_flag=True, **decorator_kwargs)
        @click.pass_context
        @wraps(fn)
        def wrapper(ctx, *args, **kwargs):
            if flag_keyword not in kwargs:
                # TODO: throw some exception
                sys.exit(1)
            to_print = kwargs[flag_keyword]

            if not keep_flag_keyword:
                del kwargs[flag_keyword]

            # Pass everything to the inner function
            output = ctx.invoke(fn, *args, **kwargs)

            # If flagged to print output, print it.
            if to_print:
                print(output)
            return output

        return wrapper
    return decorator



import math
import re
import sys

import click

from . import click_helpers


@click.group()
@click.pass_context
def fix(ctx):
    """Fixing data issues from PDF parsing mistakes."""
    pass


@fix.command()
@click.option('--in-place', default=False, is_flag=True, help="Modify the file in-place.")
@click_helpers.do_each(outer_keyword='files', inner_keyword='filename')
@click_helpers.print_output('--print_filename', default=True, help="Print the result CSV filename to STDOUT.")
@click_helpers.csv_to_dataframe(outer_keyword="filename", inner_keyword="df", keep_outer_keyword=True)
@click_helpers.dataframe_to_csv(input_keyword="filename")
@click.pass_context
def empty_rows(ctx, df):
    """Treat rows with only 1 cell to be a parser mistake. Concat that cell content to the one above it."""
    rows_to_drop = []

    for i, row in df.iterrows():
        num_empty_field = 0
        non_empty_cell_key = -1

        for key, value in row.iteritems():
            if type(value) == float and math.isnan(value):
                num_empty_field += 1
            else:
                non_empty_cell_key = key

        # If the row has only 1 field defined
        if num_empty_field == len(row) - 1 and i > 0:
            df.at[i-1, non_empty_cell_key] += df.at[i, non_empty_cell_key]
            rows_to_drop.append(i)

    df = df.drop(rows_to_drop)
    return df


@fix.command()
@click.option('--column', type=int, help="To treat the column with empty value. Begins with 0. Required.")
@click_helpers.do_each(outer_keyword='files', inner_keyword='filename')
@click_helpers.print_output('--print_filename', default=True, help="Print the result CSV filename to STDOUT.")
@click_helpers.csv_to_dataframe(outer_keyword="filename", inner_keyword="df", keep_outer_keyword=True)
@click_helpers.dataframe_to_csv(input_keyword="filename")
@click.pass_context
def column_underflow(ctx, column, df):
    """Fix the empty values which content spilled over to the last column."""
    keys = df.keys()
    accepted_values = list(filter(lambda a: type(a) == str, df[keys[column]].unique()))

    for i, row in df.iterrows():
        if type(row.iat[column]) == float and math.isnan(row.iat[column]):
            for accepted_value in accepted_values:
                if row.iat[column-1].endswith(accepted_value):
                    df.iat[i, column-1] = row.iat[column-1][:-len(accepted_value)]
                    df.iat[i, column] = accepted_value
    return df


@fix.command()
@click.option('--column', type=int, help="To treat the column with empty value. Begins with 0. Required.")
@click.option('--date-format', type=str, default="DD/MM/YYYY", help="Specify the format of the date string. Default: DD/MM/YYYY.")
@click_helpers.do_each(outer_keyword='files', inner_keyword='filename')
@click_helpers.print_output('--print_filename', default=True, help="Print the result CSV filename to STDOUT.")
@click_helpers.csv_to_dataframe(outer_keyword="filename", inner_keyword="df", keep_outer_keyword=True)
@click_helpers.dataframe_to_csv(input_keyword="filename")
@click.pass_context
def date_column_underflow(ctx, column, date_format, df):
    """Fix the empty values of date column, which content spilled over to the last column."""

    # parse date format into regex
    pattern = re.escape(date_format).upper()
    pattern = pattern.replace("YYYY", "\\d{4}")
    pattern = pattern.replace("YY", "\\d{2}")
    pattern = pattern.replace("MM", "\\d{2}")
    pattern = pattern.replace("DD", "\\d{2}")
    pattern = "^(.+?)({0})$".format(pattern)

    for i, row in df.iterrows():
        if type(row.iat[column]) == float and math.isnan(row.iat[column]):
            match = re.match(pattern, str(row.iat[column-1]))
            if match is not None:
                df.iat[i, column-1] = match.group(1)
                df.iat[i, column] = match.group(2)
    return df



from __future__ import print_function

import os
import os.path
import pickle
import sys

import click
from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build

from . import click_helpers


def __get_credentials(scopes, credentials_file="credentials.json", pickle_file="token.pickle"):
    """Get credentials for access"""
    creds = None
    # The file token.pickle stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    if os.path.exists('token.pickle'):
        with open('token.pickle', 'rb') as token:
            creds = pickle.load(token)
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(
                'credentials.json', scopes)
            #creds = flow.run_local_server(port=0)
            creds = flow.run_local_server(port=3000)
        # Save the credentials for the next run
        with open('token.pickle', 'wb') as token:
            pickle.dump(creds, token)

    return creds


def __overwrite_sheet(creds, values=[], **kwargs):
    """
    WIP: script to overwrite sheet range into.
    """
    #  pylint: disable=no-member

    service = build('sheets', 'v4', credentials=creds)
    body = {
        'values': values,
    }
    service.spreadsheets().values().update(
        body=body,
        **kwargs,
    ).execute()


def __append_sheet(creds, spreadsheet_id, range_name, appending_values):
    """
    WIP: script to overwrite sheet range into.
    """
    #  pylint: disable=no-member

    #################### read current values

    service = build('sheets', 'v4', credentials=creds)
    result = service.spreadsheets().values().get(
            spreadsheetId=spreadsheet_id,
            range=range_name,
    ).execute()

    if not result:
        print('No data found.')
    else:
        for row in result['values']:
            # Print columns A and E, which correspond to indices 0 and 4.
            print(row)

    #################### write new values

    values = result['values']
    values.append(*appending_values)
    body = {
        'values': values,
    }
    result = service.spreadsheets().values().update(
        spreadsheetId=spreadsheet_id, range=range_name,
        valueInputOption='RAW', body=body).execute()


@click.group()
@click.option('-c', '--credentials', 'credentials_file', type=str, default="credentials.json", help="The filename of the Google API credential json. Default: credentials.json")
@click.option('-p', '--pickle', 'pickle_file', type=str, default="token.pickle", help="The filename of the cookie pickle to store access token. Default: token.pickle")
@click.argument('googlesheet_id')
@click.pass_context
def googlesheet(ctx, credentials_file, pickle_file, googlesheet_id):
    """Manipulate of Google Sheet data"""

    # TODO: check if credentials_file exists
    # TODO: prompt user how to generate the credentials JSON file form API console

    ctx.obj["googlesheet_id"] = googlesheet_id
    ctx.obj["credentials"] = __get_credentials(
        [
            'https://www.googleapis.com/auth/drive.metadata.readonly',
            'https://www.googleapis.com/auth/spreadsheets',
        ],
        credentials_file=credentials_file,
        pickle_file=pickle_file,
    )


@googlesheet.command()
@click.option('-h', '--include-headers', is_flag=True, default=False, help="Include the header label as first row of the data.")
@click.option('-r', '--range', 'range_name', type=str, required=True, help="Specify the range to update.")
@click_helpers.do_each('files', 'filename')
@click_helpers.print_output('--print-filename', 'print_filename', help="Print the name of the input file to STDOUT")
@click_helpers.csv_to_dataframe('filename', keep_outer_keyword=True)
@click.pass_context
def update(ctx, filename,  df, include_headers, range_name):
    """Update google sheet"""

    # convert dataframe into 2-dimensional slice
    values = list(map(lambda a: a.tolist(), df.values))

    if include_headers:
        new_values = [list(df.keys())]
        new_values.extend(values)
        values = new_values

    __overwrite_sheet(
        ctx.obj["credentials"],
        values=values,
        spreadsheetId=ctx.obj["googlesheet_id"],
        range=range_name,
        valueInputOption='RAW',
    )

    return filename



import setuptools
import os, re, sys

with open("README.md", "r") as fh:
    long_description = fh.read()

long_description = re.sub(
    r'\[(.+?)\]\((?!(http|https):\/\/)(.+?)\)',
    r'[\1](https://gitlab.com/yookoala/scraparser/-/blob/master/\3)',
    long_description
)

name = os.getenv("PACKAGE_NAME")
if name is None:
    name = "scraparser"

version = os.getenv("VERSION")
if version is None:
    version = "0.1.0"

setuptools.setup(
    name=name,
    version=version,
    scripts=[],
    author="Koala Yeung",
    author_email="koalay@gmail.com",
    description="A simplified PDF table scraping and parsing tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/yookoala/scraparser",
    project_urls={
        "Code": "https://gitlab.com/yookoala/scraparser",
        "Issue tracker": "https://gitlab.com/yookoala/scraparser/-/issues",
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "camelot-py>=0.7"
        "click>=7.1"
        "google-api-python-client>=1.8"
        "google-auth-httplib2>=0.0.3"
        "google-auth-oauthlib>=0.4"
        "opencv-python>=4.2"
        "python-magic>=0.4"
        "validators>=0.14"
    ],
    python_requires=">=3.6",
 )



import os
import sys
import urllib.request
from datetime import datetime
from urllib.parse import urlparse

import click
import validators

from . import click_helpers


def __scrap_url(url, filename, mkdir=True, headers={}):
    if os.path.exists(filename):
        raise Exception("File already exists: {0}".format(filename))

    req = urllib.request.Request(
        url,
        data=None,
        headers=headers,
    )
    response = urllib.request.urlopen(req)
    with open(filename, "wb") as file:
        file.write(response.read())


@click.command()
@click_helpers.do_each('urls', 'url_input')
@click_helpers.print_output('--print-filename', default=True, help="Print the result filename to STDOUT.")
@click.pass_context
def scrap(ctx, url_input):
    """Scrap any URL, add some timestamp string, then save to the "./data" folder of the install location."""

    if not validators.url(url_input):
        raise Exception("Not a valid URL: {0}".format(url_input))

    url = urlparse(url_input)
    # Note: might support other scheme such as "file://" in the future,
    # but focus on HTTP and FTP for now.
    if url.scheme not in ["http", "https", "ftp"]:
        raise Exception("Not a supported URL scheme: {0}".format(url_input))

    # Rewrite the filename (from path) to add timestamp
    filename_basename = os.path.basename(url.path)
    timestring = datetime.now().strftime("%Y-%m-%d-%H%M%S")
    dirname = os.path.join(
        os.getcwd(),
        "data",
    )
    filename_scraped = os.path.join(
        dirname,
        "{0}.{1}{2}".format(
            os.path.splitext(filename_basename)[0],
            timestring,
            os.path.splitext(filename_basename)[1],
        ),
    )

    # Do the scrapping
    __scrap_url(
        url.geturl(),
        filename_scraped,
        mkdir=True,
        headers={
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.1916.47 Safari/537.36'
        },
    )
    return filename_scraped
